/*
 * Copyright (c) 2014 AsyncHttpClient Project. All rights reserved.
 *
 * This program is licensed to you under the Apache License Version 2.0,
 * and you may not use this file except in compliance with the Apache License Version 2.0.
 * You may obtain a copy of the Apache License Version 2.0 at
 *     http://www.apache.org/licenses/LICENSE-2.0.
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the Apache License Version 2.0 is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Apache License Version 2.0 for the specific language governing permissions and limitations there under.
 */
package org.asynchttpclient.netty.request;

import io.netty.channel.ChannelProgressiveFuture;
import io.netty.channel.ChannelProgressiveFutureListener;
import org.asynchttpclient.netty.NettyResponseFuture;

public class WriteProgressListener extends WriteListener implements ChannelProgressiveFutureListener {

  private final long expectedTotal;
  private long lastProgress = 0L;

  public WriteProgressListener(NettyResponseFuture<?> future,
                               boolean notifyHeaders,
                               long expectedTotal) {
    super(future, notifyHeaders);
    this.expectedTotal = expectedTotal;
  }

  @Override
  public void operationComplete(ChannelProgressiveFuture cf) {
    operationComplete(cf.channel(), cf.cause());
  }

  @Override
  public void operationProgressed(ChannelProgressiveFuture f, long progress, long total) {
    future.touch();

    if (progressAsyncHandler != null && !notifyHeaders) {
      long lastLastProgress = lastProgress;
      lastProgress = progress;
      if (total < 0) {
        total = expectedTotal;
      }
      if (progress != lastLastProgress) {
        progressAsyncHandler.onContentWriteProgress(progress - lastLastProgress, progress, total);
      }
    }
  }
}
